from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import BitAxeMake


class Max(BitAxeMake):
    raw_model = MinerModel.BITAXE.BM1397

    expected_hashboards = 1
    expected_chips = 1
    expected_fans = 1



from typing import List, Optional

from pyasic import APIError, MinerConfig
from pyasic.data import AlgoHashRate, Fan, HashBoard, HashUnit
from pyasic.device import MinerFirmware
from pyasic.miners.base import BaseMiner
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, WebAPICommand
from pyasic.web.bitaxe import BitAxeWebAPI

BITAXE_DATA_LOC = DataLocations(
    **{
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [WebAPICommand("web_system_info", "system/info")],
        ),
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_system_info", "system/info")],
        ),
    }
)


class BitAxe(BaseMiner):
    """Handler for BitAxe"""

    web: BitAxeWebAPI
    _web_cls = BitAxeWebAPI

    firmware = MinerFirmware.STOCK

    data_locations = BITAXE_DATA_LOC

    async def reboot(self) -> bool:
        await self.web.restart()
        return True

    async def get_config(self) -> MinerConfig:
        web_system_info = await self.web.system_info()
        return MinerConfig.from_bitaxe(web_system_info)

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        await self.web.update_settings(**config.as_bitaxe())

    async def _get_wattage(self, web_system_info: dict = None) -> Optional[int]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass
        if web_system_info is not None:
            try:
                return round(web_system_info["power"])
            except KeyError:
                pass

    async def _get_hashrate(
        self, web_system_info: dict = None
    ) -> Optional[AlgoHashRate]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return AlgoHashRate.SHA256(
                    web_system_info["hashRate"], HashUnit.SHA256.GH
                ).into(self.algo.unit.default)
            except KeyError:
                pass

    async def _get_uptime(self, web_system_info: dict = None) -> Optional[int]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return web_system_info["uptimeSeconds"]
            except KeyError:
                pass

    async def _get_hashboards(self, web_system_info: dict = None) -> List[HashBoard]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return [
                    HashBoard(
                        hashrate=AlgoHashRate.SHA256(
                            web_system_info["hashRate"], HashUnit.SHA256.GH
                        ).into(self.algo.unit.default),
                        chip_temp=web_system_info.get("temp"),
                        temp=web_system_info.get("vrTemp"),
                        chips=web_system_info.get("asicCount", 1),
                        expected_chips=self.expected_chips,
                        missing=False,
                        active=True,
                        voltage=web_system_info.get("voltage"),
                    )
                ]
            except KeyError:
                pass
        return []

    async def _get_fans(self, web_system_info: dict = None) -> List[Fan]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return [Fan(speed=web_system_info["fanrpm"])]
            except KeyError:
                pass
        return []

    async def _get_hostname(self, web_system_info: dict = None) -> Optional[str]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return web_system_info["hostname"]
            except KeyError:
                pass

    async def _get_api_ver(self, web_system_info: dict = None) -> Optional[str]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return web_system_info["version"]
            except KeyError:
                pass

    async def _get_fw_ver(self, web_system_info: dict = None) -> Optional[str]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return web_system_info["version"]
            except KeyError:
                pass

    async def _get_mac(self, web_system_info: dict = None) -> Optional[str]:
        if web_system_info is None:
            try:
                web_system_info = await self.web.system_info()
            except APIError:
                pass

        if web_system_info is not None:
            try:
                return web_system_info["macAddr"].upper()
            except KeyError:
                pass



# ------------------------------------------------------------------------------
#  Copyright 2024 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from __future__ import annotations

import asyncio
import warnings
from typing import Any

import httpx

from pyasic import settings
from pyasic.errors import APIError
from pyasic.web.base import BaseWebAPI


class IceRiverWebAPI(BaseWebAPI):
    def __init__(self, ip: str) -> None:
        super().__init__(ip)
        self.username = "admin"
        self.pwd = settings.get("default_iceriver_web_password", "12345678")

    async def multicommand(
        self, *commands: str, ignore_errors: bool = False, allow_warning: bool = True
    ) -> dict:
        tasks = {c: asyncio.create_task(getattr(self, c)()) for c in commands}
        await asyncio.gather(*[t for t in tasks.values()])
        return {t: tasks[t].result() for t in tasks}

    async def send_command(
        self,
        command: str | bytes,
        ignore_errors: bool = False,
        allow_warning: bool = True,
        privileged: bool = False,
        **parameters: Any,
    ) -> dict:
        async with httpx.AsyncClient(transport=settings.transport()) as client:
            try:
                # auth
                await client.post(
                    f"http://{self.ip}:{self.port}/user/loginpost",
                    params={"post": "6", "user": self.username, "pwd": self.pwd},
                )
            except httpx.HTTPError:
                warnings.warn(f"Could not authenticate with miner web: {self}")
            try:
                resp = await client.post(
                    f"http://{self.ip}:{self.port}/user/{command}", params=parameters
                )
                if not resp.status_code == 200:
                    if not ignore_errors:
                        raise APIError(f"Command failed: {command}")
                    warnings.warn(f"Command failed: {command}")
                return resp.json()
            except httpx.HTTPError:
                raise APIError(f"Command failed: {command}")

    async def locate(self, enable: bool):
        return await self.send_command(
            "userpanel", post="5", locate="1" if enable else "0"
        )

    async def userpanel(self):
        return await self.send_command("userpanel", post="4")



from __future__ import annotations

from dataclasses import dataclass

from pyasic.device.algorithm import MinerAlgo
from pyasic.device.algorithm.sha256 import SHA256Unit


@dataclass
class SHA256HashRate:
    rate: float
    unit: SHA256Unit = MinerAlgo.SHA256.unit.default

    def __float__(self):
        return float(self.rate)

    def __int__(self):
        return int(self.rate)

    def __repr__(self):
        return f"{self.rate} {str(self.unit)}"

    def __round__(self, n: int = None):
        return round(self.rate, n)

    def __add__(self, other: SHA256HashRate | int | float) -> SHA256HashRate:
        if isinstance(other, SHA256HashRate):
            return SHA256HashRate(self.rate + other.into(self.unit).rate, self.unit)
        return SHA256HashRate(self.rate + other, self.unit)

    def __sub__(self, other: SHA256HashRate | int | float) -> SHA256HashRate:
        if isinstance(other, SHA256HashRate):
            return SHA256HashRate(self.rate - other.into(self.unit).rate, self.unit)
        return SHA256HashRate(self.rate - other, self.unit)

    def __truediv__(self, other: SHA256HashRate | int | float):
        if isinstance(other, SHA256HashRate):
            return SHA256HashRate(self.rate / other.into(self.unit).rate, self.unit)
        return SHA256HashRate(self.rate / other, self.unit)

    def __floordiv__(self, other: SHA256HashRate | int | float):
        if isinstance(other, SHA256HashRate):
            return SHA256HashRate(self.rate // other.into(self.unit).rate, self.unit)
        return SHA256HashRate(self.rate // other, self.unit)

    def __mul__(self, other: SHA256HashRate | int | float):
        if isinstance(other, SHA256HashRate):
            return SHA256HashRate(self.rate * other.into(self.unit).rate, self.unit)
        return SHA256HashRate(self.rate * other, self.unit)

    def into(self, other: SHA256Unit) -> SHA256HashRate:
        return SHA256HashRate(
            rate=self.rate / (other.value / self.unit.value), unit=other
        )



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

import asyncio


class MinerListenerProtocol(asyncio.Protocol):
    def __init__(self):
        self.responses = {}
        self.transport = None
        self.new_miner = None

    async def get_new_miner(self):
        try:
            while self.new_miner is None:
                await asyncio.sleep(0)
            return self.new_miner
        finally:
            self.new_miner = None

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, data, _addr):
        if data == b"OK\x00\x00\x00\x00\x00\x00\x00\x00":
            return
        m = data.decode()
        if "," in m:
            ip, mac = m.split(",")
            if "/" in ip:
                ip = ip.replace("[", "").split("/")[0]
        else:
            d = m[:-1].split("MAC")
            ip = d[0][3:]
            mac = d[1][1:]

        self.new_miner = {"IP": ip, "MAC": mac.upper()}

    def connection_lost(self, _):
        pass


class MinerListener:
    def __init__(self, bind_addr: str = "0.0.0.0"):
        self.found_miners = []
        self.stop = asyncio.Event()
        self.bind_addr = bind_addr

    async def listen(self):
        loop = asyncio.get_running_loop()

        transport_14235, protocol_14235 = await loop.create_datagram_endpoint(
            MinerListenerProtocol, local_addr=(self.bind_addr, 14235)
        )
        transport_8888, protocol_8888 = await loop.create_datagram_endpoint(
            MinerListenerProtocol, local_addr=(self.bind_addr, 8888)
        )
        try:
            while not self.stop.is_set():
                tasks = [
                    asyncio.create_task(protocol_14235.get_new_miner()),
                    asyncio.create_task(protocol_8888.get_new_miner()),
                ]
                await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
                for t in tasks:
                    if t.done():
                        yield t.result()

        finally:
            transport_14235.close()
            transport_8888.close()

    async def cancel(self):
        self.stop = True



from __future__ import annotations

from enum import IntEnum


class SHA256Unit(IntEnum):
    H = 1
    KH = int(H) * 1000
    MH = int(KH) * 1000
    GH = int(MH) * 1000
    TH = int(GH) * 1000
    PH = int(TH) * 1000
    EH = int(PH) * 1000
    ZH = int(EH) * 1000

    default = TH

    def __str__(self):
        if self.value == self.H:
            return "H/s"
        if self.value == self.KH:
            return "KH/s"
        if self.value == self.MH:
            return "MH/s"
        if self.value == self.GH:
            return "GH/s"
        if self.value == self.TH:
            return "TH/s"
        if self.value == self.PH:
            return "PH/s"
        if self.value == self.EH:
            return "EH/s"
        if self.value == self.ZH:
            return "ZH/s"

    @classmethod
    def from_str(cls, value: str):
        if value == "H":
            return cls.H
        elif value == "KH":
            return cls.KH
        elif value == "MH":
            return cls.MH
        elif value == "GH":
            return cls.GH
        elif value == "TH":
            return cls.TH
        elif value == "PH":
            return cls.PH
        elif value == "EH":
            return cls.EH
        elif value == "ZH":
            return cls.ZH
        return cls.default

    def __repr__(self):
        return str(self)


# make this json serializable
class _SHA256Algo(str):
    unit = SHA256Unit

    def __repr__(self):
        return "SHA256Algo"


SHA256Algo = _SHA256Algo("SHA256")



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M30SPlusPlusV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusV10
    expected_hashboards = 4
    expected_chips = 255


class M30SPlusPlusV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusV20
    expected_hashboards = 4
    expected_chips = 255


class M30SPlusPlusVE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVE30
    expected_chips = 215


class M30SPlusPlusVE40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVE40
    expected_chips = 225


class M30SPlusPlusVE50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVE50
    expected_chips = 235


class M30SPlusPlusVF40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVF40
    expected_chips = 156


class M30SPlusPlusVG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVG30
    expected_chips = 111


class M30SPlusPlusVG40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVG40
    expected_chips = 117


class M30SPlusPlusVG50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVG50


class M30SPlusPlusVH10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH10
    expected_chips = 82


class M30SPlusPlusVH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH20
    expected_chips = 86


class M30SPlusPlusVH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH30
    expected_chips = 111


class M30SPlusPlusVH40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH40
    expected_chips = 70


class M30SPlusPlusVH50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH50
    expected_chips = 74


class M30SPlusPlusVH60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH60
    expected_chips = 78


class M30SPlusPlusVH70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH70
    expected_chips = 70


class M30SPlusPlusVH80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH80
    expected_chips = 74


class M30SPlusPlusVH90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH90
    expected_chips = 78


class M30SPlusPlusVH100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVH100
    expected_chips = 82


class M30SPlusPlusVJ20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVJ20


class M30SPlusPlusVJ30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusPlusVJ30



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M31SPlusV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV10

    expected_chips = 105


class M31SPlusV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV20

    expected_chips = 111


class M31SPlusV30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV30

    expected_chips = 117


class M31SPlusV40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV40

    expected_chips = 123


class M31SPlusV50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV50

    expected_chips = 148


class M31SPlusV60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV60

    expected_chips = 156


class M31SPlusV80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV80

    expected_chips = 129


class M31SPlusV90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV90

    expected_chips = 117


class M31SPlusV100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusV100

    expected_chips = 111


class M31SPlusVE10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE10

    expected_chips = 82


class M31SPlusVE20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE20

    expected_chips = 78


class M31SPlusVE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE30

    expected_chips = 105


class M31SPlusVE40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE40

    expected_chips = 111


class M31SPlusVE50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE50

    expected_chips = 117


class M31SPlusVE60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE60


class M31SPlusVE80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVE80


class M31SPlusVF20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVF20

    expected_chips = 66


class M31SPlusVF30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVF30


class M31SPlusVG20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVG20

    expected_chips = 66


class M31SPlusVG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M31SPlusVG30

    expected_chips = 70



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------

from typing import List, Optional

from pyasic.config import MinerConfig
from pyasic.data import AlgoHashRate, HashUnit
from pyasic.data.pools import PoolMetrics, PoolUrl
from pyasic.errors import APIError
from pyasic.miners.data import DataFunction, DataLocations, DataOptions, RPCAPICommand
from pyasic.miners.device.firmware import StockFirmware
from pyasic.rpc.cgminer import CGMinerRPCAPI

CGMINER_DATA_LOC = DataLocations(
    **{
        str(DataOptions.API_VERSION): DataFunction(
            "_get_api_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [RPCAPICommand("rpc_version", "version")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.EXPECTED_HASHRATE): DataFunction(
            "_get_expected_hashrate",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_stats", "stats")],
        ),
        str(DataOptions.POOLS): DataFunction(
            "_get_pools",
            [RPCAPICommand("rpc_pools", "pools")],
        ),
    }
)


class CGMiner(StockFirmware):
    """Base handler for CGMiner based miners"""

    _rpc_cls = CGMinerRPCAPI
    rpc: CGMinerRPCAPI

    data_locations = CGMINER_DATA_LOC

    async def get_config(self) -> MinerConfig:
        # get pool data
        try:
            pools = await self.rpc.pools()
        except APIError:
            return self.config

        self.config = MinerConfig.from_api(pools)
        return self.config

    ##################################################
    ### DATA GATHERING FUNCTIONS (get_{some_data}) ###
    ##################################################

    async def _get_api_ver(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                self.api_ver = rpc_version["VERSION"][0]["API"]
            except LookupError:
                pass

        return self.api_ver

    async def _get_fw_ver(self, rpc_version: dict = None) -> Optional[str]:
        if rpc_version is None:
            try:
                rpc_version = await self.rpc.version()
            except APIError:
                pass

        if rpc_version is not None:
            try:
                self.fw_ver = rpc_version["VERSION"][0]["CGMiner"]
            except LookupError:
                pass

        return self.fw_ver

    async def _get_hashrate(self, rpc_summary: dict = None) -> Optional[AlgoHashRate]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return AlgoHashRate.SHA256(
                    rpc_summary["SUMMARY"][0]["GHS 5s"], HashUnit.SHA256.GH
                ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_uptime(self, rpc_stats: dict = None) -> Optional[int]:
        if rpc_stats is None:
            try:
                rpc_stats = await self.rpc.stats()
            except APIError:
                pass

        if rpc_stats is not None:
            try:
                return int(rpc_stats["STATS"][1]["Elapsed"])
            except LookupError:
                pass

    async def _get_pools(self, rpc_pools: dict = None) -> List[PoolMetrics]:
        if rpc_pools is None:
            try:
                rpc_pools = await self.rpc.pools()
            except APIError:
                pass

        pools_data = []
        if rpc_pools is not None:
            try:
                pools = rpc_pools.get("POOLS", [])
                for pool_info in pools:
                    url = pool_info.get("URL")
                    pool_url = PoolUrl.from_str(url) if url else None
                    pool_data = PoolMetrics(
                        accepted=pool_info.get("Accepted"),
                        rejected=pool_info.get("Rejected"),
                        get_failures=pool_info.get("Get Failures"),
                        remote_failures=pool_info.get("Remote Failures"),
                        active=pool_info.get("Stratum Active"),
                        alive=pool_info.get("Status") == "Alive",
                        url=pool_url,
                        user=pool_info.get("User"),
                        index=pool_info.get("POOL"),
                    )
                    pools_data.append(pool_data)
            except LookupError:
                pass
        return pools_data



from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import ePICMake


class BlockMiner520i(ePICMake):
    raw_model = MinerModel.EPIC.BM520i

    expected_chips = 124
    expected_fans = 4


class BlockMiner720i(ePICMake):
    raw_model = MinerModel.EPIC.BM720i

    expected_chips = 180
    expected_fans = 4



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M50VE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VE30

    expected_hashboards = 4
    expected_chips = 255


class M50VG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VG30

    expected_chips = 156


class M50VH10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH10

    expected_chips = 86


class M50VH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH20

    expected_chips = 111


class M50VH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH30

    expected_chips = 117


class M50VH40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH40

    expected_chips = 84


class M50VH50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH50

    expected_chips = 105


class M50VH60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH60

    expected_chips = 84


class M50VH70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH70

    expected_chips = 105


class M50VH80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VH80

    expected_chips = 111


class M50VJ10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VJ10


class M50VJ20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VJ20

    expected_chips = 111


class M50VJ30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M50VJ30

    expected_chips = 117



from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import AuradineMake


class AuradineAI2500(AuradineMake):
    raw_model = MinerModel.AURADINE.AI2500

    expected_fans = 0



from dataclasses import dataclass

from pyasic.device.algorithm import MinerAlgo
from pyasic.device.firmware import MinerFirmware
from pyasic.device.makes import MinerMake
from pyasic.device.models import MinerModel


@dataclass
class DeviceInfo:
    make: MinerMake = None
    model: MinerModel = None
    firmware: MinerFirmware = None
    algo: MinerAlgo = None



from enum import Enum


class AntminerModels(str, Enum):
    D3 = "D3"
    HS3 = "HS3"
    L3Plus = "L3+"
    KA3 = "KA3"
    KS3 = "KS3"
    DR5 = "DR5"
    KS5 = "KS5"
    L7 = "L7"
    E9Pro = "E9Pro"
    S9 = "S9"
    S9i = "S9i"
    S9j = "S9j"
    T9 = "T9"
    Z15 = "Z15"
    S17 = "S17"
    S17Plus = "S17+"
    S17Pro = "S17 Pro"
    S17e = "S17e"
    T17 = "T17"
    T17Plus = "T17+"
    T17e = "T17e"
    S19 = "S19"
    S19NoPIC = "S19 No PIC"
    S19L = "S19L"
    S19Pro = "S19 Pro"
    S19j = "S19j"
    S19i = "S19i"
    S19Plus = "S19+"
    S19jNoPIC = "S19j No PIC"
    S19ProPlus = "S19 Pro+"
    S19jPro = "S19j Pro"
    S19jProNoPIC = "S19j Pro No PIC"
    S19jProPlus = "S19j Pro+"
    S19jProPlusNoPIC = "S19j Pro+ No PIC"
    S19XP = "S19 XP"
    S19a = "S19a"
    S19aPro = "S19a Pro"
    S19Hydro = "S19 Hydro"
    S19ProHydro = "S19 Pro Hydro"
    S19ProPlusHydro = "S19 Pro+ Hydro"
    S19KPro = "S19K Pro"
    S19kPro = "S19k Pro"
    S19kProNoPIC = "S19k Pro No PIC"
    T19 = "T19"
    S21 = "S21"
    S21Pro = "S21 Pro"
    T21 = "T21"

    def __str__(self):
        return self.value


class WhatsminerModels(str, Enum):
    M20V10 = "M20 V10"
    M20SV10 = "M20S V10"
    M20SV20 = "M20S V20"
    M20SV30 = "M20S V30"
    M20PV10 = "M20P V10"
    M20PV30 = "M20P V30"
    M20SPlusV30 = "M20S+ V30"
    M21V10 = "M21 V10"
    M21SV20 = "M21S V20"
    M21SV60 = "M21S V60"
    M21SV70 = "M21S V70"
    M21SPlusV20 = "M21S+ V20"
    M29V10 = "M29 V10"
    M30V10 = "M30 V10"
    M30V20 = "M30 V20"
    M30KV10 = "M30K V10"
    M30LV10 = "M30L V10"
    M30SV10 = "M30S V10"
    M30SV20 = "M30S V20"
    M30SV30 = "M30S V30"
    M30SV40 = "M30S V40"
    M30SV50 = "M30S V50"
    M30SV60 = "M30S V60"
    M30SV70 = "M30S V70"
    M30SV80 = "M30S V80"
    M30SVE10 = "M30S VE10"
    M30SVE20 = "M30S VE20"
    M30SVE30 = "M30S VE30"
    M30SVE40 = "M30S VE40"
    M30SVE50 = "M30S VE50"
    M30SVE60 = "M30S VE60"
    M30SVE70 = "M30S VE70"
    M30SVF10 = "M30S VF10"
    M30SVF20 = "M30S VF20"
    M30SVF30 = "M30S VF30"
    M30SVG10 = "M30S VG10"
    M30SVG20 = "M30S VG20"
    M30SVG30 = "M30S VG30"
    M30SVG40 = "M30S VG40"
    M30SVH10 = "M30S VH10"
    M30SVH20 = "M30S VH20"
    M30SVH30 = "M30S VH30"
    M30SVH40 = "M30S VH40"
    M30SVH50 = "M30S VH50"
    M30SVH60 = "M30S VH60"
    M30SVI20 = "M30S VI20"
    M30SPlusV10 = "M30S+ V10"
    M30SPlusV20 = "M30S+ V20"
    M30SPlusV30 = "M30S+ V30"
    M30SPlusV40 = "M30S+ V40"
    M30SPlusV50 = "M30S+ V50"
    M30SPlusV60 = "M30S+ V60"
    M30SPlusV70 = "M30S+ V70"
    M30SPlusV80 = "M30S+ V80"
    M30SPlusV90 = "M30S+ V90"
    M30SPlusV100 = "M30S+ V100"
    M30SPlusVE30 = "M30S+ VE30"
    M30SPlusVE40 = "M30S+ VE40"
    M30SPlusVE50 = "M30S+ VE50"
    M30SPlusVE60 = "M30S+ VE60"
    M30SPlusVE70 = "M30S+ VE70"
    M30SPlusVE80 = "M30S+ VE80"
    M30SPlusVE90 = "M30S+ VE90"
    M30SPlusVE100 = "M30S+ VE100"
    M30SPlusVF20 = "M30S+ VF20"
    M30SPlusVF30 = "M30S+ VF30"
    M30SPlusVG20 = "M30S+ VG20"
    M30SPlusVG30 = "M30S+ VG30"
    M30SPlusVG40 = "M30S+ VG40"
    M30SPlusVG50 = "M30S+ VG50"
    M30SPlusVG60 = "M30S+ VG60"
    M30SPlusVH10 = "M30S+ VH10"
    M30SPlusVH20 = "M30S+ VH20"
    M30SPlusVH30 = "M30S+ VH30"
    M30SPlusVH40 = "M30S+ VH40"
    M30SPlusVH50 = "M30S+ VH50"
    M30SPlusVH60 = "M30S+ VH60"
    M30SPlusPlusV10 = "M30S++ V10"
    M30SPlusPlusV20 = "M30S++ V20"
    M30SPlusPlusVE30 = "M30S++ VE30"
    M30SPlusPlusVE40 = "M30S++ VE40"
    M30SPlusPlusVE50 = "M30S++ VE50"
    M30SPlusPlusVF40 = "M30S++ VF40"
    M30SPlusPlusVG30 = "M30S++ VG30"
    M30SPlusPlusVG40 = "M30S++ VG40"
    M30SPlusPlusVG50 = "M30S++ VG50"
    M30SPlusPlusVH10 = "M30S++ VH10"
    M30SPlusPlusVH20 = "M30S++ VH20"
    M30SPlusPlusVH30 = "M30S++ VH30"
    M30SPlusPlusVH40 = "M30S++ VH40"
    M30SPlusPlusVH50 = "M30S++ VH50"
    M30SPlusPlusVH60 = "M30S++ VH60"
    M30SPlusPlusVH70 = "M30S++ VH70"
    M30SPlusPlusVH80 = "M30S++ VH80"
    M30SPlusPlusVH90 = "M30S++ VH90"
    M30SPlusPlusVH100 = "M30S++ VH100"
    M30SPlusPlusVJ20 = "M30S++ VJ20"
    M30SPlusPlusVJ30 = "M30S++ VJ30"
    M31V10 = "M31 V10"
    M31V20 = "M31 V20"
    M31HV10 = "M31H V10"
    M31HV40 = "M31H V40"
    M31LV10 = "M30L V10"
    M31SV10 = "M31S V10"
    M31SV20 = "M31S V20"
    M31SV30 = "M31S V30"
    M31SV40 = "M31S V40"
    M31SV50 = "M31S V50"
    M31SV60 = "M31S V60"
    M31SV70 = "M31S V70"
    M31SV80 = "M31S V80"
    M31SV90 = "M31S V90"
    M31SVE10 = "M31S VE10"
    M31SVE20 = "M31S VE20"
    M31SVE30 = "M31S VE30"
    M31SEV10 = "M31SE V10"
    M31SEV20 = "M31SE V20"
    M31SEV30 = "M31SE V30"
    M31SPlusV10 = "M31S+ V10"
    M31SPlusV20 = "M31S+ V20"
    M31SPlusV30 = "M31S+ V30"
    M31SPlusV40 = "M31S+ V40"
    M31SPlusV50 = "M31S+ V50"
    M31SPlusV60 = "M31S+ V60"
    M31SPlusV80 = "M31S+ V80"
    M31SPlusV90 = "M31S+ V90"
    M31SPlusV100 = "M31S+ V100"
    M31SPlusVE10 = "M31S+ VE10"
    M31SPlusVE20 = "M31S+ VE20"
    M31SPlusVE30 = "M31S+ VE30"
    M31SPlusVE40 = "M31S+ VE40"
    M31SPlusVE50 = "M31S+ VE50"
    M31SPlusVE60 = "M31S+ VE60"
    M31SPlusVE80 = "M31S+ VE80"
    M31SPlusVF20 = "M31S+ VF20"
    M31SPlusVF30 = "M31S+ VF30"
    M31SPlusVG20 = "M31S+ VG20"
    M31SPlusVG30 = "M31S+ VG30"
    M32V10 = "M32 V10"
    M32V20 = "M32 V20"
    M32S = "M32S"
    M33V10 = "M33 V10"
    M33V20 = "M33 V20"
    M33V30 = "M33 V30"
    M33SVG30 = "M33S VG30"
    M33SPlusVG20 = "M33S+ VG20"
    M33SPlusVH20 = "M33S+ VH20"
    M33SPlusVH30 = "M33S+ VH30"
    M33SPlusPlusVH20 = "M33S++ VH20"
    M33SPlusPlusVH30 = "M33S++ VH30"
    M33SPlusPlusVG40 = "M33S++ VG40"
    M34SPlusVE10 = "M34S+ VE10"
    M36SVE10 = "M36S VE10"
    M36SPlusVG30 = "M36S+ VG30"
    M36SPlusPlusVH30 = "M36S++ VH30"
    M39V10 = "M39 V10"
    M39V20 = "M39 V20"
    M39V30 = "M39 V30"
    M50VE30 = "M50 VE30"
    M50VG30 = "M50 VG30"
    M50VH10 = "M50 VH10"
    M50VH20 = "M50 VH20"
    M50VH30 = "M50 VH30"
    M50VH40 = "M50 VH40"
    M50VH50 = "M50 VH50"
    M50VH60 = "M50 VH60"
    M50VH70 = "M50 VH70"
    M50VH80 = "M50 VH80"
    M50VJ10 = "M50 VJ10"
    M50VJ20 = "M50 VJ20"
    M50VJ30 = "M50 VJ30"
    M50SVJ10 = "M50S VJ10"
    M50SVJ20 = "M50S VJ20"
    M50SVJ30 = "M50S VJ30"
    M50SVH10 = "M50S VH10"
    M50SVH20 = "M50S VH20"
    M50SVH30 = "M50S VH30"
    M50SVH40 = "M50S VH40"
    M50SVH50 = "M50S VH50"
    M50SPlusVH30 = "M50S+ VH30"
    M50SPlusVH40 = "M50S+ VH40"
    M50SPlusVJ30 = "M50S+ VJ30"
    M50SPlusVK20 = "M50S+ VK20"
    M50SPlusPlusVK10 = "M50S++ VK10"
    M50SPlusPlusVK20 = "M50S++ VK20"
    M50SPlusPlusVK30 = "M50S++ VK30"
    M53VH30 = "M53 VH30"
    M53SVH30 = "M53S VH30"
    M53SVJ40 = "M53S VJ40"
    M53SPlusVJ30 = "M53S+ VJ30"
    M53SPlusPlusVK10 = "M53S++ VK10"
    M56VH30 = "M56 VH30"
    M56SVH30 = "M56S VH30"
    M56SPlusVJ30 = "M56S+ VJ30"
    M59VH30 = "M59 VH30"
    M60VK10 = "M60 VK10"
    M60VK20 = "M60 VK20"
    M60VK30 = "M60 VK30"
    M60VK40 = "M60 VK40"
    M60SVK10 = "M60S VK10"
    M60SVK20 = "M60S VK20"
    M60SVK30 = "M60S VK30"
    M60SVK40 = "M60S VK40"
    M63VK10 = "M63 VK10"
    M63VK20 = "M63 VK20"
    M63VK30 = "M63 VK30"
    M63SVK10 = "M63S VK10"
    M63SVK20 = "M63S VK20"
    M63SVK30 = "M63S VK30"
    M66VK20 = "M66 VK20"
    M66VK30 = "M66 VK30"
    M66SVK20 = "M66S VK20"
    M66SVK30 = "M66S VK30"
    M66SVK40 = "M66S VK40"

    def __str__(self):
        return self.value


class AvalonminerModels(str, Enum):
    Avalon721 = "Avalon 721"
    Avalon741 = "Avalon 741"
    Avalon761 = "Avalon 761"
    Avalon821 = "Avalon 821"
    Avalon841 = "Avalon 841"
    Avalon851 = "Avalon 851"
    Avalon921 = "Avalon 921"
    Avalon1026 = "Avalon 1026"
    Avalon1047 = "Avalon 1047"
    Avalon1066 = "Avalon 1066"
    Avalon1166Pro = "Avalon 1166 Pro"
    Avalon1246 = "Avalon 1246"
    AvalonNano3 = "Avalon Nano 3"

    def __str__(self):
        return self.value


class InnosiliconModels(str, Enum):
    T3HPlus = "T3H+"
    A10X = "A10X"
    A11MX = "A11MX"

    def __str__(self):
        return self.value


class GoldshellModels(str, Enum):
    CK5 = "CK5"
    HS5 = "HS5"
    KD5 = "KD5"
    KDMax = "KD Max"
    KDBoxII = "KD Box II"
    KDBoxPro = "KD Box Pro"

    def __str__(self):
        return self.value


class ePICModels(str, Enum):
    BM520i = "BlockMiner 520i"
    BM720i = "BlockMiner 720i"

    def __str__(self):
        return self.value


class AuradineModels(str, Enum):
    AT1500 = "AT1500"
    AT2860 = "AT2860"
    AT2880 = "AT2880"
    AI2500 = "AI2500"
    AI3680 = "AI3680"
    AD2500 = "AD2500"
    AD3500 = "AD3500"

    def __str__(self):
        return self.value


class BitAxeModels(str, Enum):
    BM1366 = "Ultra"
    BM1368 = "Supra"
    BM1397 = "Max"

    def __str__(self):
        return self.value


class IceRiverModels(str, Enum):
    KS0 = "KS0"
    KS1 = "KS1"
    KS2 = "KS2"
    KS3 = "KS3"
    KS3L = "KS3L"
    KS3M = "KS3M"

    def __str__(self):
        return self.value


class MinerModel:
    ANTMINER = AntminerModels
    WHATSMINER = WhatsminerModels
    AVALONMINER = AvalonminerModels
    INNOSILICON = InnosiliconModels
    GOLDSHELL = GoldshellModels
    AURADINE = AuradineModels
    EPIC = ePICModels
    BITAXE = BitAxeModels
    ICERIVER = IceRiverModels



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M30SPlusV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV10

    expected_chips = 215


class M30SPlusV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV20

    expected_chips = 255


class M30SPlusV30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV30


class M30SPlusV40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV40

    expected_chips = 235


class M30SPlusV50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV50

    expected_chips = 225


class M30SPlusV60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV60

    expected_chips = 245


class M30SPlusV70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV70

    expected_chips = 235


class M30SPlusV80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV80

    expected_chips = 245


class M30SPlusV90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV90

    expected_chips = 225


class M30SPlusV100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusV100

    expected_chips = 215


class M30SPlusVE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE30

    expected_chips = 148


class M30SPlusVE40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE40

    expected_chips = 156


class M30SPlusVE50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE50

    expected_chips = 164


class M30SPlusVE60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE60

    expected_chips = 172


class M30SPlusVE70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE70


class M30SPlusVE80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE80


class M30SPlusVE90(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE90


class M30SPlusVE100(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVE100


class M30SPlusVF20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVF20

    expected_chips = 111


class M30SPlusVF30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVF30

    expected_chips = 117


class M30SPlusVG20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG20

    expected_chips = 82


class M30SPlusVG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG30

    expected_chips = 78


class M30SPlusVG40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG40

    expected_chips = 105


class M30SPlusVG50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG50

    expected_chips = 111


class M30SPlusVG60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVG60

    expected_chips = 86


class M30SPlusVH10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH10

    expected_chips = 64


class M30SPlusVH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH20

    expected_chips = 66


class M30SPlusVH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH30

    expected_chips = 70


class M30SPlusVH40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH40

    expected_chips = 74


class M30SPlusVH50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH50

    expected_chips = 64


class M30SPlusVH60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SPlusVH60

    expected_chips = 66



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional
from urllib.parse import urlparse


class Scheme(Enum):
    STRATUM_V1 = "stratum+tcp"
    STRATUM_V2 = "stratum2+tcp"
    STRATUM_V1_SSL = "stratum+ssl"


@dataclass
class PoolUrl:
    scheme: Scheme
    host: str
    port: int
    pubkey: Optional[str] = None

    def __str__(self) -> str:
        if self.scheme == Scheme.STRATUM_V2 and self.pubkey:
            return f"{self.scheme.value}://{self.host}:{self.port}/{self.pubkey}"
        else:
            return f"{self.scheme.value}://{self.host}:{self.port}"

    @classmethod
    def from_str(cls, url: str) -> "PoolUrl":
        parsed_url = urlparse(url)
        if not parsed_url.scheme.strip() == "":
            scheme = Scheme(parsed_url.scheme)
        else:
            scheme = Scheme.STRATUM_V1
        host = parsed_url.hostname
        port = parsed_url.port
        pubkey = parsed_url.path.lstrip("/") if scheme == Scheme.STRATUM_V2 else None
        return cls(scheme=scheme, host=host, port=port, pubkey=pubkey)


@dataclass
class PoolMetrics:
    """A dataclass to standardize pool metrics returned from miners.
    Attributes:

    accepted: Number of accepted shares.
    rejected: Number of rejected shares.
    get_failures: Number of failures in obtaining work from the pool.
    remote_failures: Number of failures communicating with the pool server.
    active: Indicates if the miner is connected to the stratum server.
    Alive : Indicates if a pool is alive.
    url: URL of the pool.
    index: Index of the pool.
    user: Username for the pool.
    pool_rejected_percent: Percentage of rejected shares by the pool.
    pool_stale_percent: Percentage of stale shares by the pool.
    """

    url: PoolUrl
    accepted: int = None
    rejected: int = None
    get_failures: int = None
    remote_failures: int = None
    active: bool = None
    alive: bool = None
    index: int = None
    user: str = None
    pool_rejected_percent: float = field(init=False)
    pool_stale_percent: float = field(init=False)

    @property
    def pool_rejected_percent(self) -> float:  # noqa - Skip PyCharm inspection
        """Calculate and return the percentage of rejected shares"""
        return self._calculate_percentage(self.rejected, self.accepted + self.rejected)

    @pool_rejected_percent.setter
    def pool_rejected_percent(self, val):
        pass

    @property
    def pool_stale_percent(self) -> float:  # noqa - Skip PyCharm inspection
        """Calculate and return the percentage of stale shares."""
        return self._calculate_percentage(
            self.get_failures, self.accepted + self.rejected
        )

    @pool_stale_percent.setter
    def pool_stale_percent(self, val):
        pass

    @staticmethod
    def _calculate_percentage(value: int, total: int) -> float:
        """Calculate the percentage."""
        if value is None or total is None:
            return 0
        if total == 0:
            return 0
        return (value / total) * 100



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
import logging
from enum import Enum
from typing import List, Optional

from pyasic.config import MinerConfig
from pyasic.data import AlgoHashRate, Fan, HashBoard, HashUnit
from pyasic.errors import APIError
from pyasic.miners.data import (
    DataFunction,
    DataLocations,
    DataOptions,
    RPCAPICommand,
    WebAPICommand,
)
from pyasic.miners.device.firmware import StockFirmware
from pyasic.rpc.gcminer import GCMinerRPCAPI
from pyasic.web.auradine import AuradineWebAPI

AURADINE_DATA_LOC = DataLocations(
    **{
        str(DataOptions.MAC): DataFunction(
            "_get_mac",
            [WebAPICommand("web_ipreport", "ipreport")],
        ),
        str(DataOptions.FW_VERSION): DataFunction(
            "_get_fw_ver",
            [WebAPICommand("web_ipreport", "ipreport")],
        ),
        str(DataOptions.HOSTNAME): DataFunction(
            "_get_hostname",
            [WebAPICommand("web_ipreport", "ipreport")],
        ),
        str(DataOptions.HASHRATE): DataFunction(
            "_get_hashrate",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
        str(DataOptions.HASHBOARDS): DataFunction(
            "_get_hashboards",
            [
                RPCAPICommand("rpc_devs", "devs"),
                WebAPICommand("web_ipreport", "ipreport"),
            ],
        ),
        str(DataOptions.WATTAGE): DataFunction(
            "_get_wattage",
            [WebAPICommand("web_psu", "psu")],
        ),
        str(DataOptions.WATTAGE_LIMIT): DataFunction(
            "_get_wattage_limit",
            [WebAPICommand("web_mode", "mode"), WebAPICommand("web_psu", "psu")],
        ),
        str(DataOptions.FANS): DataFunction(
            "_get_fans",
            [WebAPICommand("web_fan", "fan")],
        ),
        str(DataOptions.FAULT_LIGHT): DataFunction(
            "_get_fault_light",
            [WebAPICommand("web_led", "led")],
        ),
        str(DataOptions.IS_MINING): DataFunction(
            "_is_mining",
            [WebAPICommand("web_mode", "mode")],
        ),
        str(DataOptions.UPTIME): DataFunction(
            "_get_uptime",
            [RPCAPICommand("rpc_summary", "summary")],
        ),
    }
)


class AuradineLEDColors(Enum):
    OFF = 0
    GREEN = 1
    RED = 2
    YELLOW = 3
    GREEN_FLASHING = 4
    RED_FLASHING = 5
    YELLOW_FLASHING = 6

    def __int__(self):
        return self.value


class AuradineLEDCodes(Enum):
    NO_POWER = 1
    NORMAL = 2
    LOCATE_MINER = 3
    TEMPERATURE = 4
    POOL_CONFIG = 5
    NETWORK = 6
    CONTROL_BOARD = 7
    HASH_RATE_LOW = 8
    CUSTOM1 = 101
    CUSTOM2 = 102

    def __int__(self):
        return self.value


class Auradine(StockFirmware):
    """Base handler for Auradine miners"""

    _rpc_cls = GCMinerRPCAPI
    rpc: GCMinerRPCAPI
    _web_cls = AuradineWebAPI
    web: AuradineWebAPI

    data_locations = AURADINE_DATA_LOC

    supports_shutdown = True
    supports_power_modes = True
    supports_autotuning = True

    async def fault_light_on(self) -> bool:
        try:
            await self.web.set_led(code=int(AuradineLEDCodes.LOCATE_MINER))
            return True
        except APIError:
            return False

    async def fault_light_off(self) -> bool:
        try:
            await self.web.set_led(code=int(AuradineLEDCodes.NORMAL))
            return True
        except APIError:
            return False

    async def reboot(self) -> bool:
        try:
            await self.web.reboot()
        except APIError:
            return False
        return True

    async def restart_backend(self) -> bool:
        try:
            await self.web.restart_gcminer()
        except APIError:
            return False
        return True

    async def stop_mining(self) -> bool:
        try:
            await self.web.set_mode(sleep="on")
        except APIError:
            return False
        return True

    async def resume_mining(self) -> bool:
        try:
            await self.web.set_mode(sleep="off")
        except APIError:
            return False
        return True

    async def set_power_limit(self, wattage: int) -> bool:
        try:
            await self.web.set_mode(mode="custom", tune="power", power=wattage)
        except APIError:
            return False
        return True

    async def get_config(self) -> MinerConfig:
        try:
            web_conf = await self.web.multicommand("pools", "mode", "fan")
            return MinerConfig.from_auradine(web_conf=web_conf)
        except APIError as e:
            logging.warning(e)
        except LookupError:
            pass
        return MinerConfig()

    async def send_config(self, config: MinerConfig, user_suffix: str = None) -> None:
        self.config = config

        conf = config.as_auradine(user_suffix=user_suffix)
        for key in conf.keys():
            await self.web.send_command(command=key, **conf[key])

    async def upgrade_firmware(
        self,
        *,
        url: str = None,
        version: str = "latest",
        keep_settings: bool = False,
        **kwargs,
    ) -> bool:
        """
        Upgrade the firmware of the Auradine device.

        Args:
            url (str): The URL to download the firmware from.
            version (str): The version of the firmware to upgrade to.
            keep_settings (bool): Whether to keep the current settings during the upgrade.

        Returns:
            bool: True if the firmware upgrade was successful, False otherwise.
        """
        try:
            logging.info("Starting firmware upgrade process.")

            if not url and not version:
                raise ValueError(
                    "Either URL or version must be provided for firmware upgrade."
                )

            if url:
                result = await self.web.firmware_upgrade(url=url)
            else:
                result = await self.web.firmware_upgrade(version=version)

            if result.get("STATUS", [{}])[0].get("STATUS") == "S":
                logging.info("Firmware upgrade process completed successfully.")
                return True
            else:
                logging.error(
                    f"Firmware upgrade failed: {result.get('error', 'Unknown error')}"
                )
                return False

        except Exception as e:
            logging.error(
                f"An error occurred during the firmware upgrade process: {str(e)}"
            )
            return False

    ##################################################
    ### DATA GATHERING FUNCTIONS (get_{some_data}) ###
    ##################################################

    async def _get_mac(self, web_ipreport: dict = None) -> Optional[str]:
        if web_ipreport is None:
            try:
                web_ipreport = await self.web.ipreport()
            except APIError:
                pass

        if web_ipreport is not None:
            try:
                return web_ipreport["IPReport"][0]["mac"].upper()
            except (LookupError, AttributeError):
                pass

    async def _get_fw_ver(self, web_ipreport: dict = None) -> Optional[str]:
        if web_ipreport is None:
            try:
                web_ipreport = await self.web.ipreport()
            except APIError:
                pass

        if web_ipreport is not None:
            try:
                return web_ipreport["IPReport"][0]["version"]
            except LookupError:
                pass

    async def _get_hostname(self, web_ipreport: dict = None) -> Optional[str]:
        if web_ipreport is None:
            try:
                web_ipreport = await self.web.ipreport()
            except APIError:
                pass

        if web_ipreport is not None:
            try:
                return web_ipreport["IPReport"][0]["hostname"]
            except LookupError:
                pass

    async def _get_hashrate(self, rpc_summary: dict = None) -> Optional[AlgoHashRate]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return AlgoHashRate.SHA256(
                    rpc_summary["SUMMARY"][0]["MHS 5s"], HashUnit.SHA256.MH
                ).into(self.algo.unit.default)
            except (LookupError, ValueError, TypeError):
                pass

    async def _get_hashboards(
        self, rpc_devs: dict = None, web_ipreport: dict = None
    ) -> List[HashBoard]:
        hashboards = [
            HashBoard(slot=i, expected_chips=self.expected_chips)
            for i in range(self.expected_hashboards)
        ]

        if rpc_devs is None:
            try:
                rpc_devs = await self.rpc.devs()
            except APIError:
                pass
        if web_ipreport is None:
            try:
                web_ipreport = await self.web.ipreport()
            except APIError:
                pass

        if rpc_devs is not None:
            try:
                for board in rpc_devs["DEVS"]:
                    b_id = board["ID"] - 1
                    hashboards[b_id].hashrate = AlgoHashRate.SHA256(
                        board["MHS 5s"], HashUnit.SHA256.MH
                    ).into(self.algo.unit.default)
                    hashboards[b_id].temp = round(float(float(board["Temperature"])), 2)
                    hashboards[b_id].missing = False
            except LookupError:
                pass

        if web_ipreport is not None:
            try:
                for board, sn in enumerate(web_ipreport["IPReport"][0]["HBSerialNo"]):
                    hashboards[board].serial_number = sn
                    hashboards[board].missing = False
            except LookupError:
                pass

        return hashboards

    async def _get_wattage(self, web_psu: dict = None) -> Optional[int]:
        if web_psu is None:
            try:
                web_psu = await self.web.get_psu()
            except APIError:
                pass

        if web_psu is not None:
            try:
                return int(float(web_psu["PSU"][0]["PowerIn"].replace("W", "")))
            except (LookupError, TypeError, ValueError):
                pass

    async def _get_wattage_limit(
        self, web_mode: dict = None, web_psu: dict = None
    ) -> Optional[int]:
        if web_mode is None:
            try:
                web_mode = await self.web.get_mode()
            except APIError:
                pass

        if web_mode is not None:
            try:
                return web_mode["Mode"][0]["Power"]
            except (LookupError, TypeError, ValueError):
                pass

        if web_psu is None:
            try:
                web_psu = await self.web.get_psu()
            except APIError:
                pass

        if web_psu is not None:
            try:
                return int(float(web_psu["PSU"][0]["PoutMax"].replace("W", "")))
            except (LookupError, TypeError, ValueError):
                pass

    async def _get_fans(self, web_fan: dict = None) -> List[Fan]:
        if web_fan is None:
            try:
                web_fan = await self.web.get_fan()
            except APIError:
                pass

        fans = []
        if web_fan is not None:
            try:
                for fan in web_fan["Fan"]:
                    fans.append(Fan(round(fan["Speed"])))
            except LookupError:
                pass
        return fans

    async def _get_fault_light(self, web_led: dict = None) -> Optional[bool]:
        if web_led is None:
            try:
                web_led = await self.web.get_led()
            except APIError:
                pass

        if web_led is not None:
            try:
                return web_led["LED"][0]["Code"] == int(AuradineLEDCodes.LOCATE_MINER)
            except LookupError:
                pass

    async def _is_mining(self, web_mode: dict = None) -> Optional[bool]:
        if web_mode is None:
            try:
                web_mode = await self.web.get_mode()
            except APIError:
                pass

        if web_mode is not None:
            try:
                return web_mode["Mode"][0]["Sleep"] == "off"
            except (LookupError, TypeError, ValueError):
                pass

    async def _get_uptime(self, rpc_summary: dict = None) -> Optional[int]:
        if rpc_summary is None:
            try:
                rpc_summary = await self.rpc.summary()
            except APIError:
                pass

        if rpc_summary is not None:
            try:
                return rpc_summary["SUMMARY"][0]["Elapsed"]
            except LookupError:
                pass



from __future__ import annotations

import asyncio
import json
from typing import Any

import httpx

from pyasic import settings
from pyasic.web.base import BaseWebAPI


class MaraWebAPI(BaseWebAPI):
    def __init__(self, ip: str) -> None:
        super().__init__(ip)

    async def multicommand(
        self, *commands: str, ignore_errors: bool = False, allow_warning: bool = True
    ) -> dict:
        async with httpx.AsyncClient(transport=settings.transport()) as client:
            tasks = [
                asyncio.create_task(self._handle_multicommand(client, command))
                for command in commands
            ]
            all_data = await asyncio.gather(*tasks)

        data = {}
        for item in all_data:
            data.update(item)

        data["multicommand"] = True
        return data

    async def _handle_multicommand(
        self, client: httpx.AsyncClient, command: str
    ) -> dict:
        auth = httpx.DigestAuth(self.username, self.pwd)

        try:
            url = f"http://{self.ip}:{self.port}/kaonsu/v1/{command}"
            ret = await client.get(url, auth=auth)
        except httpx.HTTPError:
            pass
        else:
            if ret.status_code == 200:
                try:
                    json_data = ret.json()
                    return {command: json_data}
                except json.decoder.JSONDecodeError:
                    pass
        return {command: {}}

    async def send_command(
        self,
        command: str | bytes,
        ignore_errors: bool = False,
        allow_warning: bool = True,
        privileged: bool = False,
        **parameters: Any,
    ) -> dict:
        url = f"http://{self.ip}:{self.port}/kaonsu/v1/{command}"
        auth = httpx.DigestAuth(self.username, self.pwd)
        try:
            async with httpx.AsyncClient(
                transport=settings.transport(),
            ) as client:
                if parameters:
                    data = await client.post(
                        url,
                        auth=auth,
                        timeout=settings.get("api_function_timeout", 3),
                        json=parameters,
                    )
                else:
                    data = await client.get(url, auth=auth)
        except httpx.HTTPError:
            pass
        else:
            if data.status_code == 200:
                try:
                    return data.json()
                except json.decoder.JSONDecodeError:
                    pass

    async def brief(self):
        return await self.send_command("brief")

    async def ping(self):
        return await self.send_command("ping")

    async def get_locate_miner(self):
        return await self.send_command("locate_miner")

    async def set_locate_miner(self, blinking: bool):
        return await self.send_command("locate_miner", blinking=blinking)

    async def reboot(self):
        return await self.send_command("maintenance", type="reboot")

    async def reset(self):
        return await self.send_command("maintenance", type="reset")

    async def reload(self):
        return await self.send_command("maintenance", type="reload")

    async def set_password(self, new_pwd: str):
        return await self.send_command(
            "maintenance",
            type="passwd",
            params={"curPwd": self.pwd, "confirmPwd": self.pwd, "newPwd": new_pwd},
        )

    async def get_network_config(self):
        return await self.send_command("network_config")

    async def set_network_config(self, **params):
        return await self.send_command("network_config", **params)

    async def get_miner_config(self):
        return await self.send_command("miner_config")

    async def set_miner_config(self, **params):
        return await self.send_command("miner_config", **params)

    async def fans(self):
        return await self.send_command("fans")

    async def log(self):
        return await self.send_command("log")

    async def overview(self):
        return await self.send_command("overview")

    async def connections(self):
        return await self.send_command("connections")

    async def controlboard_info(self):
        return await self.send_command("controlboard_info")

    async def event_chart(self):
        return await self.send_command("event_chart")

    async def hashboards(self):
        return await self.send_command("hashboards")

    async def pools(self):
        return await self.send_command("pools")



# ------------------------------------------------------------------------------
#  Copyright 2022 Upstream Data Inc                                            -
#                                                                              -
#  Licensed under the Apache License, Version 2.0 (the "License");             -
#  you may not use this file except in compliance with the License.            -
#  You may obtain a copy of the License at                                     -
#                                                                              -
#      http://www.apache.org/licenses/LICENSE-2.0                              -
#                                                                              -
#  Unless required by applicable law or agreed to in writing, software         -
#  distributed under the License is distributed on an "AS IS" BASIS,           -
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    -
#  See the License for the specific language governing permissions and         -
#  limitations under the License.                                              -
# ------------------------------------------------------------------------------
from pyasic.device.models import MinerModel
from pyasic.miners.device.makes import WhatsMinerMake


class M30SV10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV10

    expected_chips = 148


class M30SV20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV20

    expected_chips = 156


class M30SV30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV30

    expected_chips = 164


class M30SV40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV40

    expected_chips = 172


class M30SV50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV50

    expected_chips = 156


class M30SV60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV60

    expected_chips = 164


class M30SV70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV70


class M30SV80(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SV80

    expected_chips = 129


class M30SVE10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE10

    expected_chips = 105


class M30SVE20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE20

    expected_chips = 111


class M30SVE30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE30

    expected_chips = 117


class M30SVE40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE40

    expected_chips = 123


class M30SVE50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE50

    expected_chips = 129


class M30SVE60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE60


class M30SVE70(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVE70


class M30SVF10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVF10

    expected_chips = 70


class M30SVF20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVF20

    expected_chips = 74


class M30SVF30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVF30

    expected_chips = 78


class M30SVG10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVG10

    expected_chips = 66


class M30SVG20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVG20

    expected_chips = 70


class M30SVG30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVG30

    expected_chips = 74


class M30SVG40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVG40

    expected_chips = 78


class M30SVH10(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVH10

    expected_chips = 64


class M30SVH20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVH20

    expected_chips = 66


class M30SVH30(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVH30


class M30SVH40(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVH40

    expected_chips = 64


class M30SVH50(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVH50

    expected_chips = 66


class M30SVH60(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVH60


class M30SVI20(WhatsMinerMake):
    raw_model = MinerModel.WHATSMINER.M30SVI20

    expected_chips = 70



from enum import Enum

from pyasic.data.hashrate.sha256 import SHA256HashRate
from pyasic.device.algorithm.sha256 import SHA256Unit


class AlgoHashRate(Enum):
    SHA256 = SHA256HashRate

    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)


class HashUnit:
    SHA256 = SHA256Unit
